;; Generated package description from gnorb.el
(define-package "gnorb" "1.0.1" "Glue code between Gnus, Org, and BBDB" '((cl-lib "0.5")) :url "https://github.com/girzel/gnorb" :keywords '("mail" "org" "gnus" "bbdb" "todo" "task"))
